/**
 * PETALS - PETALS Services Platform. Copyright (c) 2005 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id: Pojo.java 676 2006-06-27 15:44:03Z alouis $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.se.pojo;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.logging.Logger;

import javax.jbi.component.ComponentContext;
import javax.jbi.messaging.DeliveryChannel;

import org.ow2.petals.component.framework.api.exception.FaultException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.process.async.AsyncContext;
import org.ow2.petals.se.pojo.su.POJOURLClassLoader;

public class Pojo {

    public static final String COMPONENT_CONTEXT_SETTER = "setComponentContext";

    public static final String DELIVERY_CHANNEL_SETTER = "setDeliveryChannel";

    public static final String JBI_LISTENER_SETTER = "setJBIListener";

    public static final String LOGGER_SETTER = "setLogger";

    public static final String ON_EXCHANGE_METHOD = "onExchange";

    public static final String ON_ASYNC_EXCHANGE_METHOD = "onAsyncExchange";

    private DeliveryChannel channel;

    private POJOURLClassLoader classLoader;

    private ComponentContext context;

    private Logger logger;

    private Method onExchangeMethod;

    private Method onAsyncExchangeMethod;

    private Object pojo;

    public Pojo(final Object object, final POJOURLClassLoader classLoader,
            final ComponentContext context, final DeliveryChannel channel, final Logger logger) {
        this.pojo = object;
        this.context = context;
        this.channel = channel;
        this.logger = logger;
        this.classLoader = classLoader;
    }

    public Object getPojo() {
        return this.pojo;
    }

    /**
     * Call the Exchange handler method of the POJO
     * 
     * @param ex
     * @return true if the OUT message has been set on the exchange and has to
     *         be send
     * @throws Exception
     */
    public boolean callOnExchangeMethod(Exchange ex, AsyncContext asyncContext) throws Exception {

        final Method onExchange = (asyncContext == null ? onExchangeMethod : onAsyncExchangeMethod);

        // First, we set the SU classloader
        ClassLoader oldClassLoader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(this.classLoader);

        boolean result = false;
        try {
            Object o = (asyncContext == null ? onExchange.invoke(pojo, ex) : onExchange.invoke(
                    pojo, ex, asyncContext));

            result = (Boolean) o;

        } catch (final IllegalArgumentException e) {
            throw new Exception("Java reflection exception during call on " + onExchange.getName()
                    + "() on the POJO.", e);
        } catch (final IllegalAccessException e) {
            throw new Exception("Java reflection exception during call on " + onExchange.getName()
                    + "() on the POJO.", e);
        } catch (final InvocationTargetException e) {
            // An exception is thrown during the processing of the POJO class.
            // This exception will be convert in a fault or an ERROR ack
            final Throwable orginalEx = e.getCause();
            if (orginalEx != null && orginalEx instanceof FaultException) {
                throw (FaultException) orginalEx;
            } else {
                throw new Exception("Processing exception during call on " + onExchange.getName()
                        + "() on the POJO.", e);
            }
        } finally {
            // We restore the thread classloader
            Thread.currentThread().setContextClassLoader(oldClassLoader);
        }
        return result;
    }

    public ClassLoader getClassLoader() {
        return this.classLoader;
    }

    public void init() throws Exception {
        if (pojo == null) {
            throw new Exception("POJO is null.");
        }
        setLoggerOnPojo();
        setContextOnPojo();
        setChannelOnPojo();

        call("init");

        setupOnExchangeMethod();
        setupOnAsyncExchangeMethod();
    }

    /**
     * TODO: Use when the POJO would be bound to the SU life-cycle.
     * 
     * @throws Exception
     */
    public void shutdown() throws Exception {
        if (pojo == null) {
            throw new Exception("POJO is null.");
        }
        this.classLoader.close(this.logger);
        call("shutdown");
    }

    /**
     * TODO: Use when the POJO would be bound to the SU life-cycle.
     * 
     * @throws Exception
     */
    public void start() throws Exception {
        if (pojo == null) {
            throw new Exception("POJO is null.");
        }
        call("start");
    }

    /**
     * TODO: Use when the POJO would be bound to the SU life-cycle.
     * 
     * @throws Exception
     */
    public void stop() throws Exception {
        if (pojo == null) {
            throw new Exception("POJO is null.");
        }
        call("stop");
    }

    /**
     * Call the method on the pojo. If the method is not found, nothing happens.
     * The method can not have parameters
     * 
     * @throws PEtALSComponentSDKException
     *             invocation failed or the method has thrown an exception
     */
    protected void call(String methodName) throws Exception {

        // First, we set the SU classloader
        ClassLoader oldClassLoader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(this.classLoader);

        try {
            Method m = findMethod(pojo.getClass(), methodName, null);

            if (m != null) {
                try {
                    m.invoke(pojo);
                } catch (IllegalArgumentException e) {
                    throw new Exception("java reflection exception during call on " + methodName
                            + "() on the POJO.", e);
                } catch (IllegalAccessException e) {
                    throw new Exception("java reflection exception during call on " + methodName
                            + "() on the POJO.", e);
                } catch (InvocationTargetException e) {
                    // An exception is thrown calling the POJO class method.
                    throw new Exception("processing exception during call on " + methodName
                            + "() on the POJO.", e);
                } catch (Throwable e) {
                    throw new Exception(e.getMessage(), e);
                }

            }
        } finally {
            // We restore the thread classloader
            Thread.currentThread().setContextClassLoader(oldClassLoader);
        }
    }

    /**
     * Find a "set...(DeliveryChannel channel)" on the pojo and set the
     * DeliveryChannel on it. The method is not mandatory.
     * 
     * @throws PEtALSComponentSDKException
     *             invocation failed
     */
    protected void setChannelOnPojo() throws Exception {

        // First, we set the SU classloader
        ClassLoader oldClassLoader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(this.classLoader);

        try {
            Method setChannelMethod = findMethod(pojo.getClass(), DELIVERY_CHANNEL_SETTER,
                    DeliveryChannel.class);

            if (setChannelMethod != null) {
                try {
                    setChannelMethod.invoke(pojo, channel);

                } catch (Throwable e) {
                    throw new Exception("Can not set the DeliveryChannel on the POJO.", e);
                }
            }
        } finally {
            // We restore the thread classloader
            Thread.currentThread().setContextClassLoader(oldClassLoader);
        }
    }

    /**
     * Find a "setComponentContext(ComponentContext context)" on the POJO and
     * set the ComponentContext on it. The method is not mandatory.
     * 
     * @throws PEtALSComponentSDKException
     *             invocation failed
     */
    protected void setContextOnPojo() throws Exception {

        // First, we set the SU classloader
        ClassLoader oldClassLoader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(this.classLoader);

        try {
            Method setContextMethod = findMethod(pojo.getClass(), COMPONENT_CONTEXT_SETTER,
                    ComponentContext.class);

            if (setContextMethod != null) {
                try {
                    setContextMethod.invoke(pojo, context);

                } catch (Throwable e) {
                    throw new Exception("Can not set the ComponentContext on the POJO.", e);
                }
            }
        } finally {
            // We restore the thread classloader
            Thread.currentThread().setContextClassLoader(oldClassLoader);
        }
    }

    /**
     * Find a "set...(Logger logger)" on the pojo and set the logger on it. The
     * method is not mandatory.
     * 
     * @throws PEtALSComponentSDKException
     *             invocation failed
     */
    protected void setLoggerOnPojo() throws Exception {

        // First, we set the SU classloader
        ClassLoader oldClassLoader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(this.classLoader);

        try {
            Method setLoggerMethod = findMethod(pojo.getClass(), LOGGER_SETTER, Logger.class);

            if (setLoggerMethod != null) {
                try {
                    setLoggerMethod.invoke(pojo, logger);

                } catch (Throwable e) {
                    throw new Exception("Can not set the Logger on the POJO.", e);
                }
            }
        } finally {
            // We restore the thread classloader
            Thread.currentThread().setContextClassLoader(oldClassLoader);
        }
    }

    /**
     * Find a "on...(Exchange exchange)" on the pojo to be used when message
     * exchanges are accepted.
     * 
     * @throws PEtALSComponentSDKException
     *             method not found or invocation failed
     */
    protected void setupOnExchangeMethod() throws Exception {

        onExchangeMethod = findMethod(pojo.getClass(), ON_EXCHANGE_METHOD, Exchange.class);

        if (onExchangeMethod == null) {
            throw new Exception("A handler method for Exchange is not found in " + pojo.getClass());

        }
        if (onExchangeMethod.getReturnType() != Boolean.TYPE) {
            throw new Exception("The handler method for Exchange must return a boolean.");
        }

    }

    /**
     * Find a "on...(Exchange exchange)" on the pojo to be used when message
     * exchanges are accepted.
     * 
     * @throws PEtALSComponentSDKException
     *             method not found or invocation failed
     */
    @SuppressWarnings("unchecked")
    protected void setupOnAsyncExchangeMethod() throws Exception {

        Class args[] = new Class[] { Exchange.class, AsyncContext.class };
        try {
            onAsyncExchangeMethod = pojo.getClass().getMethod(ON_ASYNC_EXCHANGE_METHOD, args);
        } catch (NoSuchMethodException e) {
            onAsyncExchangeMethod = null; // Async method is optional
        }

        if (onAsyncExchangeMethod != null && onAsyncExchangeMethod.getReturnType() != Boolean.TYPE) {
            throw new Exception("The handler method for Async Exchange must return a boolean.");
        }

    }

    /**
     * Find a method on the specified class (including those inherited from
     * superclasses) with the specified name, and having the optional specified
     * class or super-class as parameter. TODO; handle method return type
     * 
     * @param clazz
     *            class on which the method has to be found
     * @param name
     *            the method name
     * @param parameterClass
     *            parameter of the class (only one parameter is allow, but is
     *            not mandatory)
     * @return the matching method, or null if no one match
     */
    public static final Method findMethod(Class<?> clazz, String name, Class<?> parameterClass) {
        Method method = null;
        Method[] ms = clazz.getMethods();

        for (int i = 0; i < ms.length && method == null; i++) {
            Method tmpMethod = ms[i];
            if (tmpMethod.getName().equals(name) && Modifier.isPublic(tmpMethod.getModifiers())) {
                if (parameterClass != null && tmpMethod.getParameterTypes().length == 1
                        && parameterClass.isAssignableFrom(tmpMethod.getParameterTypes()[0])) {
                    method = tmpMethod;
                } else if (parameterClass == null && tmpMethod.getParameterTypes().length == 0) {
                    method = tmpMethod;
                }
            }
        }

        return method;
    }
}
